# vim: set fileencoding=utf-8 :
#
# Copyright (C) 2012 Daniyal Kazempour
# University of Tuebingen
# Interfaculty Institute of Microbiology and Infection Medicine
# Div. of Microbiology/Biotechnology
#
# Copyright (C) 2012 Kai Blin
# University of Tuebingen
# Interfaculty Institute of Microbiology and Infection Medicine
# Div. of Microbiology/Biotechnology
#
# License: GNU Affero General Public License v3 or later
# A copy of GNU AGPL v3 should have been included in this software package in LICENSE.txt.
'''
More detailed lantipeptide analysis using HMMer-based leader peptide
cleavage site prediction as well as prediction of number of lanthionine
bridges and molcular mass.
'''

import logging
from Bio.SeqFeature import SeqFeature, FeatureLocation
from Bio.SeqUtils.ProtParam import ProteinAnalysis
from antismash import utils

known_precursor_domains = (
    'Antimicr18',
    'Gallidermin',
    'L_biotic_A',
    'lacticin_l',
    'leader_d',
    'leader_abc',
    'leader_eh',
    'mature_ha',
    'lacticin_mat',
    'mature_b',
    'mature_d',
    'mature_ab',
    'mature_a',
    'TIGR03731',
    'LD_lanti_pre',
    'strep_PEQAXS',
)

class Lantipeptide(object):
    '''
    Class to calculate and store lantipeptide information
    '''
    def __init__(self, start, end, score, lantype):
        self.start = start
        self.end = end
        self.score = score
        self.lantype = lantype
        self._core = ''
        self._weight = -1
        self._lan_bridges = -1

    @property
    def core(self):
        return self._core

    @core.setter
    def core(self, seq):
        self.core_analysis = ProteinAnalysis(seq, monoisotopic=True)
        self._core = seq

    def __repr__(self):
        return "Lantipeptide(%s..%s, %s, %r, %r, %s, %s)" % (self.start, self.end, self.score, self.lantype, self._core, self._lan_bridges, self._weight)

    @property
    def number_of_lan_bridges(self):
        '''
        function determines the number of lanthionine bridges in the core peptide
        '''
        if self._lan_bridges > -1:
            return self._lan_bridges

        aas = self.core_analysis.count_amino_acids()
        no_cys = aas['C']
        no_thr_ser = aas['T'] + aas['S']
        self._lan_bridges = min(no_cys, no_thr_ser)
        return self._lan_bridges

    @property
    def molecular_weight(self):
        '''
        function determines the weight of the core peptide and substracts
        the weight which is reduced, due to dehydratation
        '''
        if self._weight > -1:
            return self._weight

        if not self._core:
            raise ValueError()

        aas = self.core_analysis.count_amino_acids()
        no_thr_ser = aas['T'] + aas['S']

        mol_mass = int(round(self.core_analysis.molecular_weight()))
        mods = 18 * no_thr_ser
        self._weight = mol_mass - mods
        return self._weight


def predict_cleavage_site(query_hmmfile, target_sequence):
    '''
    Function extracts from HMMER the start position, end position, score and lant-type
    of the HMM alignment
    '''
    hmmer_res = utils.run_hmmpfam2(query_hmmfile, target_sequence)
    resvec = None
    for res in hmmer_res:
        for hits in res:
            lanti_type = hits.description
            for hsp in hits:
                resvec = Lantipeptide(hsp.query_start-1, hsp.query_end, hsp.bitscore, lanti_type)
                return resvec
    return resvec


thresh_dict = {
        'Class-I' : 4,
        'Class-II' : -3,
        'Class-III' : 5,
        'Class-IV' : 140
    }

def get_best_result(result_list):
    '''
    Function determines the lantipeptide class, according to the hsp
    it returns the result vector containing the highest score
    '''
    highest_score_vector = None
    for result in result_list:
        #get vector with highest score from all 4 lan-classes
        if highest_score_vector is None or (result.score > highest_score_vector.score):
            highest_score_vector = result

    #now perform check, if score is within the boundaries not exceeding
    #the thresholds. otherwhise it is considered as a false-positive
    if highest_score_vector is not None and \
       thresh_dict[highest_score_vector.lantype] <= highest_score_vector.score:
        return highest_score_vector

    return


def run_lantipred(seq_record, query):
    hmmer_profile_list = ['ClassILanti.hmm', 'ClassIILanti.hmm', 'ClassIIILanti.hmm', 'ClassIVLanti.hmm']
    hmmer_res_list = []

    query_sequence = utils.get_aa_sequence(seq_record, query, to_stop=True)
    lan_a_fasta = ">%s\n%s" % (utils.get_gene_id(query), query_sequence)

    #run sequence against profiles and parse them in a vector containing START, END, SCORE and LANTYPE
    for profile in hmmer_profile_list:
        profile = utils.get_full_path(__file__, profile)
        hmmer_res_list.append(predict_cleavage_site(profile, lan_a_fasta))

    #get the best-match, gained by comparing scores and checking if threshhold for respectative class is not violated
    result = get_best_result(hmmer_res_list)

    if result is None:
        logging.debug('Failed to predict lantipeptide cleavage.')
        return

    #extract now (that class is known and thus the END component) the core peptide
    result.core = query_sequence[result.end:]

    return result


def find_lan_a_features(seq_record, cluster):
    lan_a_features = []
    for feature in utils.get_cds_features(seq_record):
        if feature.location.start < cluster.location.start or \
           feature.location.end > cluster.location.end:
            continue

        if not 'sec_met' in feature.qualifiers:
            continue

        domain = None
        for entry in feature.qualifiers['sec_met']:
            if entry.startswith('Domains detected:'):
                domain = entry.split()[2]
                break

        if domain is None:
            continue

        if domain not in known_precursor_domains:
            continue

        lan_a_features.append(feature)

    return lan_a_features


def result_vec_to_features(orig_feature, res_vec):
    start = orig_feature.location.start
    end = orig_feature.location.start + (res_vec.end * 3)
    strand = orig_feature.location.strand
    loc = FeatureLocation(start, end, strand=strand)
    leader_feature = SeqFeature(loc, type='CDS_motif')
    leader_feature.qualifiers['note'] = ['leader peptide']
    leader_feature.qualifiers['locus_tag'] = [ utils.get_gene_id(orig_feature) ]

    start = end
    end = orig_feature.location.end
    loc = FeatureLocation(start, end, strand=strand)
    core_feature = SeqFeature(loc, type='CDS_motif')
    core_feature.qualifiers['note'] = ['core peptide']
    core_feature.qualifiers['note'].append('molecular weight: %s' % res_vec.molecular_weight)
    core_feature.qualifiers['note'].append('number of bridges: %s' % res_vec.number_of_lan_bridges)
    core_feature.qualifiers['note'].append('predicted core seq: %s' % res_vec.core)
    core_feature.qualifiers['note'].append('predicted class: %s' % res_vec.lantype)
    core_feature.qualifiers['locus_tag'] = [ utils.get_gene_id(orig_feature) ]

    return [leader_feature, core_feature]


def specific_analysis(seq_record, options):
    clusters = utils.get_cluster_features(seq_record)
    for cluster in clusters:
        if 'product' not in cluster.qualifiers or \
           'lantipeptide' not in cluster.qualifiers['product'][0]:
            continue

        lan_as = find_lan_a_features(seq_record, cluster)
        for lan_a in lan_as:
            result_vec = run_lantipred(seq_record, lan_a)
            if result_vec is None:
                continue
            new_features = result_vec_to_features(lan_a, result_vec)
            seq_record.features.extend(new_features)

